/**************************************************************************
This is a part of ProxyView.
Copyright (c) 2002-2003 EinStEin <einstein@underground.org.pl>.

This file is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
**************************************************************************/

#include "main.h"
#include "extern.h"

enum { SERVER, CLIENT };

int lport = 0;
unsigned char lhost[256];

int rport = 0;
unsigned char rhost[256];

int serversock, clientsock, sock;
int disconnected = FALSE;

int verbose_read = FALSE;
int verbose_write = FALSE;

int display_binary = FALSE;
int display_raw = FALSE;

int display_server = FALSE;
int display_client = FALSE;

void
handle_signal(int dummy)
{
	switch (dummy) {
	case 0:
		printf("Bup.\n");
		break;
	case SIGINT:
		printf("Ouch?! Auua :( Ctrl+C...\n");
		break;
	case SIGHUP:
		printf("Terminal sie powiesil :-)\n");
		break;
	case SIGSEGV:
		printf("\"Unibus timeout fatal trap program lost sorry\"\n");
		break;
	case SIGBUS:
		printf("Brakpamiecinawetnaspacje.\n");
		break;
	case SIGPIPE:
		printf("Ktos wyszedl w trakcie rozmowy.\n");
		break;
	default:
		printf("N.D.S.K.N.P.W. #%d\n", dummy);
	}

	close(clientsock);
	close(serversock);
	close(sock);

	exit(127);
}

void
print_proxy(const unsigned char *buffer, int type, const unsigned char *comm,
	    int mode, unsigned int size)
{
	if ((display_server == FALSE) && (type == SERVER))
		return;
	if ((display_client == FALSE) && (type == CLIENT))
		return;

	if ((mode == FALSE) && (size != 0)) {
		int pointer = 0;

		if (!display_raw)
			printf(comm);

		for (pointer = 0; pointer < size; pointer++) {
			if (display_binary) {
				printf("%#.2x ", buffer[pointer]);
			} else {
				if (buffer[pointer] == '\r') {
					if (display_raw)
						printf("\r");
					else
						printf("<CR>");
				} else if (buffer[pointer] == '\n') {
					if (display_raw)
						printf("\n");
					else
						printf("<LF>");
				} else {
					printf("%c", buffer[pointer]);
				}
			}
		}

		if (!display_raw)
			printf("\n");
	}
	fflush(NULL);
}

int
readfromhost(int sock, unsigned char *buffer, int type,
	     const unsigned char *comm, unsigned int *size)
{
	fd_set rfds;
	struct timeval tv;
	int retval;

	FD_ZERO(&rfds);
	FD_SET(sock, &rfds);

	tv.tv_sec = 0;
	tv.tv_usec = 1;

	retval = select(sock + 1, &rfds, NULL, NULL, &tv);

	if (retval) {

		if ((*size = recv(sock, buffer, __READ_BUF_SIZE, 0)) <= 0)
			disconnected = TRUE;

		print_proxy(buffer, type, comm, verbose_read, *size);

		return 0;

	} else
		return -1;
}

int
writetohost(int sock, const unsigned char *buffer, int type,
	    const unsigned char *comm, unsigned int *size)
{
	fd_set wfds;
	struct timeval tv;
	int retval;

	if (size <= 0)
		return 0;

	FD_ZERO(&wfds);
	FD_SET(sock, &wfds);

	tv.tv_sec = 0;
	tv.tv_usec = 1;

	retval = select(sock + 1, NULL, &wfds, NULL, &tv);

	if (retval) {

		print_proxy(buffer, type, comm, verbose_write, *size);

		if (send(sock, buffer, *size, 0) > 0) {
			return (0);
		} else {
			disconnected = TRUE;
			return -1;
		}

	}

	return -2;
}

int
main(int argc, char *argv[])
{
	int c = 0;
	unsigned int size = 0;
	struct sockaddr_in addr;
	struct hostent *hp = 0;

	signal(SIGSEGV, handle_signal);
	signal(SIGINT, handle_signal);
	signal(SIGHUP, handle_signal);
	signal(SIGBUS, handle_signal);
	signal(SIGPIPE, handle_signal);
	signal(SIGTERM, handle_signal);
	signal(SIGILL, handle_signal);
	signal(SIGFPE, handle_signal);
	signal(SIGCLD, SIG_IGN);

	while (1) {
		int option_index = 0;
		static struct option long_options[] = {
			{"bind", 1, 0, 'b'},
			{"host", 1, 0, 'h'},
			{"display-server", 0, 0, 's'},
			{"display-client", 0, 0, 'c'},
			{"verbose-read", 0, 0, 'r'},
			{"verbose-write", 0, 0, 'w'},
			{"binary", 0, 0, 'n'},
			{"raw", 0, 0, 'a'},
			{"help", 0, 0, '?'},
			{"version", 0, 0, 'V'},
			{0, 0, 0, 0}
		};

		c = getopt_long(argc, argv, "b:h:scrwna?V", long_options,
				&option_index);

		if (c == -1)
			break;

		switch (c) {
		case 'b':
			parse_hostname_and_port(optarg, lhost, &lport);
			break;

		case 'h':
			parse_hostname_and_port(optarg, rhost, &rport);
			break;

		case 's':
			display_server = TRUE;
			break;

		case 'c':
			display_client = TRUE;
			break;

		case 'r':
			verbose_read = TRUE;
			break;

		case 'w':
			verbose_write = TRUE;
			break;

		case 'n':
			display_binary = TRUE;
			break;

		case 'a':
			display_raw = TRUE;
			break;

		case '?':
			usage(argv[0]);
			exit(EXIT_SUCCESS);
			break;

		case 'V':
			version();
			exit(EXIT_SUCCESS);
			break;

		default:
			break;
		}
	}

	if (!lport || !rport) {
		printf("%s: no remote or local port!\n", argv[0]);
		exit(1);
	}

	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		perror("socket()");
		exit(1);
	} else
		printf("socket(): ok\n");

	memset(&addr, 0, sizeof (addr));
	addr.sin_family = AF_INET;

	if (strlen(lhost)) {
		if ((hp = gethostbyname(lhost)) == 0) {
			addr.sin_addr.s_addr = inet_addr(lhost);
			if ((int) addr.sin_addr.s_addr == -1)
				perror("gethostbyname()");
			else
				printf("gethostbyname(): ok\n");
		} else
			memcpy((char *) &addr.sin_addr.s_addr, hp->h_addr,
			       hp->h_length);
	} else {
		addr.sin_addr.s_addr = htonl(INADDR_ANY);
	}

	addr.sin_port = htons(lport);

	if ((bind(sock, (struct sockaddr *) (&addr), sizeof (addr))) != 0) {
		perror("bind()");
		exit(2);
	} else
		printf("bind(): ok\n");

	if ((listen(sock, __LISTQ)) != 0) {
		perror("listen()");
		exit(3);
	} else
		printf("listen(): ok\n");

	if (!strlen(rhost)) {
		sprintf(rhost, "localhost");
	}

	if (!strlen(lhost)) {
		sprintf(lhost, "<ANY>");
	}

	printf("\nProxyView configuration:\n");
	printf("Server side      : %s:%i [%s]\n", rhost, rport,
	       display_server ? "display data" : "no data display");
	printf("Client side      : %s:%i [%s]\n", lhost, lport,
	       display_client ? "display data" : "no data display");
	printf("Verbose read     : %s\n", verbose_read ? "TRUE" : "FALSE");
	printf("Verbose write    : %s\n", verbose_write ? "TRUE" : "FALSE");
	printf("Binary dump mode : %s\n", display_binary ? "TRUE" : "FALSE");
	printf("RAW dump mode    : %s\n\n", display_raw ? "TRUE" : "FALSE");

	for (;;) {
		printf("*** waiting for client on %i\n", lport);
		serversock = accept(sock, NULL, NULL);
		disconnected = FALSE;

		struct sockaddr_in peerAddr;
		socklen_t peerLen = sizeof (peerAddr);
		getpeername(serversock, (struct sockaddr *) &peerAddr,
			    &peerLen);

		printf("\n*** client connected to %i, from %s\n\n", lport,
		       inet_ntoa(peerAddr.sin_addr));

		if ((hp = gethostbyname(rhost)) == 0) {
			addr.sin_addr.s_addr = inet_addr(rhost);
			if ((int) addr.sin_addr.s_addr == -1)
				perror("gethostbyname()");
			else
				printf("gethostbyname(): ok\n");
		} else
			memcpy((char *) &addr.sin_addr.s_addr, hp->h_addr,
			       hp->h_length);

		if ((clientsock = socket(AF_INET, SOCK_STREAM, 0)) == -1)
			perror("socket()");
		else
			printf("socket(): ok\n");

		addr.sin_family = AF_INET;
		addr.sin_port = htons(rport);
		setsockopt(clientsock, SOL_SOCKET, SO_LINGER, 0, 0);
		setsockopt(clientsock, SOL_SOCKET, SO_KEEPALIVE, 0, 0);

		if (connect
		    (clientsock, (struct sockaddr *) &addr,
		     sizeof (addr)) == -1)
			perror("connect()");
		else
			printf("connect(): ok\n");

		printf("\n*** started proxy to %s:%i\n\n", rhost, rport);

		for (;;) {

			unsigned char tmp[4096];

			if (disconnected) {
				close(clientsock);
				close(serversock);
				break;
			}

			size = 0;
			memset(tmp, 0, 4096);
			readfromhost(serversock, tmp, SERVER, "C>S (read): ",
				     &size);
			if (size)
				writetohost(clientsock, tmp, CLIENT,
					    "C>S (write): ", &size);

			size = 0;
			memset(tmp, 0, 4096);
			readfromhost(clientsock, tmp, CLIENT, "S>C (read): ",
				     &size);
			if (size)
				writetohost(serversock, tmp, SERVER,
					    "S>C (write): ", &size);
		}

		printf
		    ("\n*** ended proxy [ client or server closed socket ]\n\n");
	}

	close(sock);
}
